@@ -15,20 +15,22 @@
  * Non-empty lists have a header, which will not be relocated as long as the
  * list remains non-empty, and an expansible data array.
  *
- * We support three types of lists:
+ * We support four types of lists:
  *
  *	T_List: lists of pointers
  *		(in practice usually pointers to Nodes, but not always;
  *		declared as "void *" to minimize casting annoyances)
  *	T_IntList: lists of integers
  *	T_OidList: lists of Oids
+ *	T_XidList: lists of TransactionIds
+ *		(the XidList infrastructure is less complete than the other cases)
  *
- * (At the moment, ints and Oids are the same size, but they may not
- * always be so; try to be careful to maintain the distinction.)
+ * (At the moment, ints, Oids, and XIDs are the same size, but they may not
+ * always be so; be careful to use the appropriate list type for your data.)
  *
  *
- * Portions Copyright (c) 2003-2020, PgPool Global Development Group
- * Portions Copyright (c) 1996-2020, PostgreSQL Global Development Group
+ * Portions Copyright (c) 2003-2023, PgPool Global Development Group
+ * Portions Copyright (c) 1996-2023, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * src/include/nodes/pg_list.h
@@ -42,16 +44,23 @@
 #include "pool_parser.h"
 #include "nodes.h"
 
+/*
+ *  * include/c.h
+ *   */
+typedef uint32 TransactionId;
+
+
 typedef union ListCell
 {
 	void	   *ptr_value;
 	int			int_value;
 	Oid			oid_value;
+	TransactionId xid_value;
 } ListCell;
 
 typedef struct List
 {
-	NodeTag		type;			/* T_List, T_IntList, or T_OidList */
+	NodeTag		type;			/* T_List, T_IntList, T_OidList, or T_XidList */
 	int			length;			/* number of elements currently present */
 	int			max_length;		/* allocated length of elements[] */
 	ListCell   *elements;		/* re-allocatable array of cells */
@@ -146,26 +155,6 @@ list_second_cell(const List *l)
 		return NULL;
 }
 
-/* Fetch address of list's third cell, if it has one, else NULL */
-static inline ListCell *
-list_third_cell(const List *l)
-{
-	if (l && l->length >= 3)
-		return &l->elements[2];
-	else
-		return NULL;
-}
-
-/* Fetch address of list's fourth cell, if it has one, else NULL */
-static inline ListCell *
-list_fourth_cell(const List *l)
-{
-	if (l && l->length >= 4)
-		return &l->elements[3];
-	else
-		return NULL;
-}
-
 /* Fetch list's length */
 static inline int
 list_length(const List *l)
@@ -188,35 +177,36 @@ list_length(const List *l)
  * linitial() than lfirst(): given a List, lsecond() returns the data
  * in the second list cell.
  */
-
 #define lfirst(lc)				((lc)->ptr_value)
 #define lfirst_int(lc)			((lc)->int_value)
 #define lfirst_oid(lc)			((lc)->oid_value)
+#define lfirst_xid(lc)			((lc)->xid_value)
 #define lfirst_node(type,lc)	castNode(type, lfirst(lc))
 
-#define linitial(l)				lfirst(list_head(l))
-#define linitial_int(l)			lfirst_int(list_head(l))
-#define linitial_oid(l)			lfirst_oid(list_head(l))
+#define linitial(l)				lfirst(list_nth_cell(l, 0))
+#define linitial_int(l)			lfirst_int(list_nth_cell(l, 0))
+#define linitial_oid(l)			lfirst_oid(list_nth_cell(l, 0))
 #define linitial_node(type,l)	castNode(type, linitial(l))
 
-#define lsecond(l)				lfirst(list_second_cell(l))
-#define lsecond_int(l)			lfirst_int(list_second_cell(l))
-#define lsecond_oid(l)			lfirst_oid(list_second_cell(l))
+#define lsecond(l)				lfirst(list_nth_cell(l, 1))
+#define lsecond_int(l)			lfirst_int(list_nth_cell(l, 1))
+#define lsecond_oid(l)			lfirst_oid(list_nth_cell(l, 1))
 #define lsecond_node(type,l)	castNode(type, lsecond(l))
 
-#define lthird(l)				lfirst(list_third_cell(l))
-#define lthird_int(l)			lfirst_int(list_third_cell(l))
-#define lthird_oid(l)			lfirst_oid(list_third_cell(l))
+#define lthird(l)				lfirst(list_nth_cell(l, 2))
+#define lthird_int(l)			lfirst_int(list_nth_cell(l, 2))
+#define lthird_oid(l)			lfirst_oid(list_nth_cell(l, 2))
 #define lthird_node(type,l)		castNode(type, lthird(l))
 
-#define lfourth(l)				lfirst(list_fourth_cell(l))
-#define lfourth_int(l)			lfirst_int(list_fourth_cell(l))
-#define lfourth_oid(l)			lfirst_oid(list_fourth_cell(l))
+#define lfourth(l)				lfirst(list_nth_cell(l, 3))
+#define lfourth_int(l)			lfirst_int(list_nth_cell(l, 3))
+#define lfourth_oid(l)			lfirst_oid(list_nth_cell(l, 3))
 #define lfourth_node(type,l)	castNode(type, lfourth(l))
 
-#define llast(l)				lfirst(list_tail(l))
-#define llast_int(l)			lfirst_int(list_tail(l))
-#define llast_oid(l)			lfirst_oid(list_tail(l))
+#define llast(l)				lfirst(list_last_cell(l))
+#define llast_int(l)			lfirst_int(list_last_cell(l))
+#define llast_oid(l)			lfirst_oid(list_last_cell(l))
+#define llast_xid(l)			lfirst_xid(list_last_cell(l))
 #define llast_node(type,l)		castNode(type, llast(l))
 
 /*
@@ -225,6 +215,7 @@ list_length(const List *l)
 #define list_make_ptr_cell(v)	((ListCell) {.ptr_value = (v)})
 #define list_make_int_cell(v)	((ListCell) {.int_value = (v)})
 #define list_make_oid_cell(v)	((ListCell) {.oid_value = (v)})
+#define list_make_xid_cell(v)	((ListCell) {.xid_value = (v)})
 
 #define list_make1(x1) \
 	list_make1_impl(T_List, list_make_ptr_cell(x1))
@@ -236,6 +227,10 @@ list_length(const List *l)
 #define list_make4(x1,x2,x3,x4) \
 	list_make4_impl(T_List, list_make_ptr_cell(x1), list_make_ptr_cell(x2), \
 					list_make_ptr_cell(x3), list_make_ptr_cell(x4))
+#define list_make5(x1,x2,x3,x4,x5) \
+	list_make5_impl(T_List, list_make_ptr_cell(x1), list_make_ptr_cell(x2), \
+					list_make_ptr_cell(x3), list_make_ptr_cell(x4), \
+					list_make_ptr_cell(x5))
 
 #define list_make1_int(x1) \
 	list_make1_impl(T_IntList, list_make_int_cell(x1))
@@ -247,6 +242,10 @@ list_length(const List *l)
 #define list_make4_int(x1,x2,x3,x4) \
 	list_make4_impl(T_IntList, list_make_int_cell(x1), list_make_int_cell(x2), \
 					list_make_int_cell(x3), list_make_int_cell(x4))
+#define list_make5_int(x1,x2,x3,x4,x5) \
+	list_make5_impl(T_IntList, list_make_int_cell(x1), list_make_int_cell(x2), \
+					list_make_int_cell(x3), list_make_int_cell(x4), \
+					list_make_int_cell(x5))
 
 #define list_make1_oid(x1) \
 	list_make1_impl(T_OidList, list_make_oid_cell(x1))
@@ -258,6 +257,25 @@ list_length(const List *l)
 #define list_make4_oid(x1,x2,x3,x4) \
 	list_make4_impl(T_OidList, list_make_oid_cell(x1), list_make_oid_cell(x2), \
 					list_make_oid_cell(x3), list_make_oid_cell(x4))
+#define list_make5_oid(x1,x2,x3,x4,x5) \
+	list_make5_impl(T_OidList, list_make_oid_cell(x1), list_make_oid_cell(x2), \
+					list_make_oid_cell(x3), list_make_oid_cell(x4), \
+					list_make_oid_cell(x5))
+
+#define list_make1_xid(x1) \
+	list_make1_impl(T_XidList, list_make_xid_cell(x1))
+#define list_make2_xid(x1,x2) \
+	list_make2_impl(T_XidList, list_make_xid_cell(x1), list_make_xid_cell(x2))
+#define list_make3_xid(x1,x2,x3) \
+	list_make3_impl(T_XidList, list_make_xid_cell(x1), list_make_xid_cell(x2), \
+					list_make_xid_cell(x3))
+#define list_make4_xid(x1,x2,x3,x4) \
+	list_make4_impl(T_XidList, list_make_xid_cell(x1), list_make_xid_cell(x2), \
+					list_make_xid_cell(x3), list_make_xid_cell(x4))
+#define list_make5_xid(x1,x2,x3,x4,x5) \
+	list_make5_impl(T_XidList, list_make_xid_cell(x1), list_make_xid_cell(x2), \
+					list_make_xid_cell(x3), list_make_xid_cell(x4), \
+					list_make_xid_cell(x5))
 
 /*
  * Locate the n'th cell (counting from 0) of the list.
@@ -271,6 +289,16 @@ list_nth_cell(const List *list, int n)
 	return &list->elements[n];
 }
 
+/*
+ * Return the last cell in a non-NIL List.
+ */
+static inline ListCell *
+list_last_cell(const List *list)
+{
+	Assert(list != NIL);
+	return &list->elements[list->length - 1];
+}
+
 /*
  * Return the pointer value contained in the n'th element of the
  * specified list. (List elements begin at 0.)
@@ -382,6 +410,32 @@ lnext(const List *l, const ListCell *c)
  */
 #define foreach_current_index(cell)  (cell##__state.i)
 
+/*
+ * for_each_from -
+ *	  Like foreach(), but start from the N'th (zero-based) list element,
+ *	  not necessarily the first one.
+ *
+ * It's okay for N to exceed the list length, but not for it to be negative.
+ *
+ * The caveats for foreach() apply equally here.
+ */
+#define for_each_from(cell, lst, N)	\
+	for (ForEachState cell##__state = for_each_from_setup(lst, N); \
+		 (cell##__state.l != NIL && \
+		  cell##__state.i < cell##__state.l->length) ? \
+		 (cell = &cell##__state.l->elements[cell##__state.i], true) : \
+		 (cell = NULL, false); \
+		 cell##__state.i++)
+
+static inline ForEachState
+for_each_from_setup(const List *lst, int N)
+{
+	ForEachState r = {lst, N};
+
+	Assert(N >= 0);
+	return r;
+}
+
 /*
  * for_each_cell -
  *	  a convenience macro which loops through a list starting from a
@@ -398,7 +452,7 @@ lnext(const List *l, const ListCell *c)
 		 cell##__state.i++)
 
 static inline ForEachState
-for_each_cell_setup(List *lst, ListCell *initcell)
+for_each_cell_setup(const List *lst, const ListCell *initcell)
 {
 	ForEachState r = {lst,
 	initcell ? list_cell_number(lst, initcell) : list_length(lst)};
@@ -449,8 +503,8 @@ for_each_cell_setup(List *lst, ListCell *initcell)
 		 cell1##__state.i1++, cell1##__state.i2++)
 
 static inline ForBothCellState
-for_both_cell_setup(List *list1, ListCell *initcell1,
-					List *list2, ListCell *initcell2)
+for_both_cell_setup(const List *list1, const ListCell *initcell1,
+					const List *list2, const ListCell *initcell2)
 {
 	ForBothCellState r = {list1, list2,
 		initcell1 ? list_cell_number(list1, initcell1) : list_length(list1),
@@ -507,37 +561,57 @@ extern List *list_make3_impl(NodeTag t, ListCell datum1, ListCell datum2,
 							 ListCell datum3);
 extern List *list_make4_impl(NodeTag t, ListCell datum1, ListCell datum2,
 							 ListCell datum3, ListCell datum4);
+extern List *list_make5_impl(NodeTag t, ListCell datum1, ListCell datum2,
+							 ListCell datum3, ListCell datum4,
+							 ListCell datum5);
 
-extern List *lappend(List *list, void *datum);
-extern List *lappend_int(List *list, int datum);
-extern List *lappend_oid(List *list, Oid datum);
+/*
+ * from PostgreSQL src/include/c.h
+ *
+ * pg_nodiscard means the compiler should warn if the result of a function
+ * call is ignored.  The name "nodiscard" is chosen in alignment with
+ * (possibly future) C and C++ standards.  For maximum compatibility, use it
+ * as a function declaration specifier, so it goes before the return type.
+ */
+#ifdef __GNUC__
+#define pg_nodiscard __attribute__((warn_unused_result))
+#else
+#define pg_nodiscard
+#endif
 
-extern List *list_insert_nth(List *list, int pos, void *datum);
-extern List *list_insert_nth_int(List *list, int pos, int datum);
-extern List *list_insert_nth_oid(List *list, int pos, Oid datum);
+extern pg_nodiscard List *lappend(List *list, void *datum);
+extern pg_nodiscard List *lappend_int(List *list, int datum);
+extern pg_nodiscard List *lappend_oid(List *list, Oid datum);
+extern pg_nodiscard List *lappend_xid(List *list, TransactionId datum);
 
-extern List *lcons(void *datum, List *list);
-extern List *lcons_int(int datum, List *list);
-extern List *lcons_oid(Oid datum, List *list);
+extern pg_nodiscard List *list_insert_nth(List *list, int pos, void *datum);
+extern pg_nodiscard List *list_insert_nth_int(List *list, int pos, int datum);
+extern pg_nodiscard List *list_insert_nth_oid(List *list, int pos, Oid datum);
 
-extern List *list_concat(List *list1, const List *list2);
-extern List *list_concat_copy(const List *list1, const List *list2);
+extern pg_nodiscard List *lcons(void *datum, List *list);
+extern pg_nodiscard List *lcons_int(int datum, List *list);
+extern pg_nodiscard List *lcons_oid(Oid datum, List *list);
 
-extern List *list_truncate(List *list, int new_size);
+extern pg_nodiscard List *list_concat(List *list1, const List *list2);
+extern pg_nodiscard List *list_concat_copy(const List *list1, const List *list2);
+
+extern pg_nodiscard List *list_truncate(List *list, int new_size);
 
 extern bool list_member(const List *list, const void *datum);
 extern bool list_member_ptr(const List *list, const void *datum);
 extern bool list_member_int(const List *list, int datum);
 extern bool list_member_oid(const List *list, Oid datum);
+extern bool list_member_xid(const List *list, TransactionId datum);
 
-extern List *list_delete(List *list, void *datum);
-extern List *list_delete_ptr(List *list, void *datum);
-extern List *list_delete_int(List *list, int datum);
-extern List *list_delete_oid(List *list, Oid datum);
-extern List *list_delete_first(List *list);
-extern List *list_delete_last(List *list);
-extern List *list_delete_nth_cell(List *list, int n);
-extern List *list_delete_cell(List *list, ListCell *cell);
+extern pg_nodiscard List *list_delete(List *list, void *datum);
+extern pg_nodiscard List *list_delete_ptr(List *list, void *datum);
+extern pg_nodiscard List *list_delete_int(List *list, int datum);
+extern pg_nodiscard List *list_delete_oid(List *list, Oid datum);
+extern pg_nodiscard List *list_delete_first(List *list);
+extern pg_nodiscard List *list_delete_last(List *list);
+extern pg_nodiscard List *list_delete_first_n(List *list, int n);
+extern pg_nodiscard List *list_delete_nth_cell(List *list, int n);
+extern pg_nodiscard List *list_delete_cell(List *list, ListCell *cell);
 
 extern List *list_union(const List *list1, const List *list2);
 extern List *list_union_ptr(const List *list1, const List *list2);
@@ -554,28 +628,32 @@ extern List *list_difference_ptr(const List *list1, const List *list2);
 extern List *list_difference_int(const List *list1, const List *list2);
 extern List *list_difference_oid(const List *list1, const List *list2);
 
-extern List *list_append_unique(List *list, void *datum);
-extern List *list_append_unique_ptr(List *list, void *datum);
-extern List *list_append_unique_int(List *list, int datum);
-extern List *list_append_unique_oid(List *list, Oid datum);
+extern pg_nodiscard List *list_append_unique(List *list, void *datum);
+extern pg_nodiscard List *list_append_unique_ptr(List *list, void *datum);
+extern pg_nodiscard List *list_append_unique_int(List *list, int datum);
+extern pg_nodiscard List *list_append_unique_oid(List *list, Oid datum);
 
-extern List *list_concat_unique(List *list1, const List *list2);
-extern List *list_concat_unique_ptr(List *list1, const List *list2);
-extern List *list_concat_unique_int(List *list1, const List *list2);
-extern List *list_concat_unique_oid(List *list1, const List *list2);
+extern pg_nodiscard List *list_concat_unique(List *list1, const List *list2);
+extern pg_nodiscard List *list_concat_unique_ptr(List *list1, const List *list2);
+extern pg_nodiscard List *list_concat_unique_int(List *list1, const List *list2);
+extern pg_nodiscard List *list_concat_unique_oid(List *list1, const List *list2);
 
 extern void list_deduplicate_oid(List *list);
 
 extern void list_free(List *list);
 extern void list_free_deep(List *list);
 
-extern List *list_copy(const List *list);
-extern List *list_copy_tail(const List *list, int nskip);
-extern List *list_copy_deep(const List *oldlist);
+extern pg_nodiscard List *list_copy(const List *oldlist);
+extern pg_nodiscard List *list_copy_head(const List *oldlist, int len);
+extern pg_nodiscard List *list_copy_tail(const List *oldlist, int nskip);
+extern pg_nodiscard List *list_copy_deep(const List *oldlist);
 
 typedef int (*list_sort_comparator) (const ListCell *a, const ListCell *b);
 extern void list_sort(List *list, list_sort_comparator cmp);
 
+extern int	list_int_cmp(const ListCell *p1, const ListCell *p2);
 extern int	list_oid_cmp(const ListCell *p1, const ListCell *p2);
 
+extern char *NameListToString(List *names);
+
 #endif							/* PG_LIST_H */
